@@ -420,38 +420,35 @@ module Init = struct
     stop.depends <- a :: stop.depends;
     a
 
-  let exec a =
+  let rec exec a =
     let log =
       if conf_trace#get then fun s ->
         let id = Thread.id (Thread.self ()) in
         Printf.printf "init(%i):%-35s@%s\n%!" id a.name s
       else fun _ -> ()
     in
-    let rec exec a =
-      log "called";
-      Mutex.lock a.mutex;
-      try
-        if not a.launched then begin
-          a.launched <- true;
-          log "start";
-          log "start-depends";
-          List.iter exec a.depends;
-          log "stop-depends";
-          log "start-atom";
-          a.f ();
-          log "stop-atom";
-          log "start-triggers";
-          List.iter exec a.triggers;
-          log "stop-triggers";
-          log "stop"
-        end;
-        Mutex.unlock a.mutex;
-        log "return"
-      with e ->
-        Mutex.unlock a.mutex;
-        raise e
-    in
-    exec a
+    log "called";
+    Mutex.lock a.mutex;
+    try
+      if not a.launched then begin
+        a.launched <- true;
+        log "start";
+        log "start-depends";
+        List.iter exec a.depends;
+        log "stop-depends";
+        log "start-atom";
+        a.f ();
+        log "stop-atom";
+        log "start-triggers";
+        List.iter exec a.triggers;
+        log "stop-triggers";
+        log "stop"
+      end;
+      Mutex.unlock a.mutex;
+      log "return"
+    with e ->
+      Mutex.unlock a.mutex;
+      raise e
 
   let rec wait_signal () =
     try ignore (Thread.wait_signal [Sys.sigterm; Sys.sigint]) with
@@ -462,7 +459,7 @@ module Init = struct
   exception StartError of exn
   exception StopError of exn
 
-  (* Dummy functions in the case where 
+  (* Dummy functions in the case where
    * Printexc does not have the required
    * functions. *)
   let get_backtrace () =
@@ -623,6 +620,8 @@ module Log = struct
 
   type t =
     < active : int -> bool
+    ; level : int
+    ; set_level : int -> unit
     ; path : Conf.path
     ; f : 'a. int -> ('a, unit, string, unit) format4 -> 'a
     ; g :
@@ -758,39 +757,16 @@ module Log = struct
         Queue.push entry !log_queue;
         Condition.signal log_condition)
 
-  let build path =
-    let rec aux p l (t : Conf.ut) =
-      match p with
-        | [] -> t :: l
-        | s :: q ->
-            let st =
-              try t#path [s]
-              with Conf.Unbound _ ->
-                let c = Conf.int ~p:(t#plug s) "subordinate log level" in
-                c#ut
-            in
-            aux q (t :: l) st
-    in
-    aux path [] conf_level#ut
-
   let make path : t =
-    let confs = build path in
     let path_str = Conf.string_of_path path in
+    let conf_level = ref (fun () -> conf_level#get) in
     object (self : t)
       method path = path
 
-      method active level =
-        let rec aux l =
-          match l with
-            | [] -> None
-            | t :: q -> (
-                match aux q with
-                  | Some i -> Some i
-                  | None -> (
-                      try Some (Conf.as_int t)#get
-                      with Conf.Undefined _ -> None))
-        in
-        match aux confs with None -> false | Some i -> i >= level
+      method active level = level <= !conf_level ()
+
+      method level = !conf_level ()
+      method set_level level = conf_level := (fun () -> level)
 
       method g ?(colorize = fun x -> x) level =
         match self#active level with
@@ -863,6 +839,7 @@ module Log = struct
       match !log_thread with
         | None -> ()
         | Some th ->
+            log_thread := None;
             Condition.signal log_condition;
             Thread.join th
     end;
